<?php
/***
    MSC HMIPHeatingControl für IPS
    Skript zur Bedienung von HomeMatic IP - Thermostaten und virtuellen Heizungsgruppen
    Version: 2022-02-09

    - Dieses Skript in IPS speichern und Pfad zur xmlrpc.inc anpassen
    - Für jeden Raum bzw. jede virtuell Heizungsgruppe eine Dummy-Instanz anlegen.
    -- Innerhalb der Dummy-Instanz mindestens eine HomeMatic HMIP-Heating-Instanz der virtuellen Heizungsgruppe, optional die Wartungs-Instanzen anlegen bzw. dorthin verschieben.
    -- In der Dummy-Instanz ein neues Skript anlegen, welches dieses Skript mittels Object-ID inkludiert:
        $script =  IPS_GetScript(ID dieses Skriptes);
        include($script['ScriptFile']); 
    -- Bei virtuellen Heizungsgruppen VOR dem Include-Befehl die Adresse der virtuellen Gruppe konfigurieren:
        $customConfig["HMAdressHeatingControl"] = "INT0000005:1";
    -- Optional können alle Parameter aus $config mittels $customConfig in der Instanz überschrieben werden.
    -- Das Skript innerhalb der Dummy-Instanz einmalig ausführen. Alle Variablen etc sollten angelegt werden.
***/

// Einbindung xmlrpc -> über Google oder Symcon Forum suchen und einbinden
require "phpxmlrpc-4.3.0/lib/xmlrpc.inc" ;

/*** globale Config ***/
// scheduleGroups -> Liste von Binärflags für die Zeitabschnitte des Wochenplans (Mo - Fr (1 + 2 + 4 + 8 + 16) = 31 / Sa + So (32 + 64) = 96)
$config = [
    "idHeatingControl" => null, // Wenn null (default), wird eine HomeMatic -Instanz in IPS auf aktueller Strukturebene gesucht.
    "HMAdressHeatingControl" => null, // Optional für z.B. Heizungsgruppen (z.B. INT0000001:1). Wird ansonsten aus HomeMatic -Instanz in IPS ermittelt.
    "profiles" => [
        "P1" => ["id" => 1, "name" => "A", "scheduleGroups" => [31,96]],
        "P2" => ["id" => 2, "name" => "B", "scheduleGroups" => [31,96]],
        "P3" => ["id" => 3, "name" => "C", "scheduleGroups" => [1,2,4,8,16,32,64]],
    ],
    "elements" => [
        ["name" => "Temperatur", "type" => "linkHeatingControlVariable", "sourceIdent" => "ACTUAL_TEMPERATURE"],
        ["name" => "Luftfeuchtigkeit", "type" => "linkHeatingControlVariable", "sourceIdent" => "HUMIDITY"],
        //["name" => "Ventil-Öffnungsgrad", "type" => "linkHeatingControlVariable", "sourceIdent" => "VALVE_STATE"],
        ["name" => "Boost", "type" => "linkHeatingControlVariable", "sourceIdent" => "BOOST_MODE", "profile" => "~Switch", "selfAsCustomAction" => true],
        ["name" => "Heizmodus", "type" => "heatingModeCombined", "profile" => "MSC.HMIPHeatingControl.HeatingModeCombined", "selfAsCustomAction" => true],
        ["name" => "Modus", "type" => "setHeatingControlProfile", "sourceIdent" => "SET_POINT_MODE", "profile" => "MSC.HMIPHeatingControl.SetPointMode", "selfAsCustomAction" => true],
        ["name" => "Solltemperatur aktueller Modus", "type" => "linkHeatingControlVariable", "sourceIdent" => "SET_POINT_TEMPERATURE", "profile" => "MSC.HMIPHeatingControl.Temperature", "selfAsCustomAction" => true],
        ["name" => "Wochenprofil", "type" => "setHeatingControlProfile", "sourceIdent" => "ACTIVE_PROFILE", "profile" => "MSC.HMIPHeatingControl.ActiveProfile", "selfAsCustomAction" => true],
        ["name" => "Wochenprofile laden", "type" => "script", "ident" => "load_profiles", "content" => "<? IPS_RunScriptEx(" . $_IPS['SELF'] . ",['targetAction' => 'loadProfiles']); ?>"],
        ["name" => "Wochenprofile speichern", "type" => "script", "ident" => "store_profiles", "content" => "<? IPS_RunScriptEx(" . $_IPS['SELF'] . ",['targetAction' => 'storeProfiles']); ?>"],
        ["name" => "Liste Wochenprofile", "type" => "profiles"], 
        ["name" => "Status", "ident" => "status", "type" => "setPosition"], 
        ["name" => "Fenster", "type" => "linkHeatingControlVariable", "sourceIdent" => "WINDOW_STATE", "profile" => "MSC.HMIPHeatingControl.Window","triggerOnChange" => "<? IPS_RunScriptEx(" . $_IPS['SELF'] . ",['targetAction' => 'toggleWindowState']); ?>"],
        ["name" => "Batterie", "type" => "linkHeatingControlMaintenanceVariables", "sourceIdent" => "LOW_BAT"],
    ],
    "overwriteExistingVariablesProfiles" => true, // den Statusvariablen zugewiesene Variablenprofile überschreiben
    "allowWindowOpenTemperatureChange" => false, // bei geöffnetem Fenster ist die Fenster-Auf-Temperatur änderbar. Die Änderung wird erst beim nächsten Öffen des Fensters in den Thermostaten wirksam.
    "lockDeviceButtonOnOpenedWindow" => false, // bei geöffnetem Fenster die globale Bediensperre für alle HM-Geräte der Gruppe aktivieren

];

// Config mit optionaler $customConfig mergen
if (isset($customConfig) && is_array($customConfig)) $config = array_merge($config,$customConfig);

// Default - Parameter für Wochenprofile setzen
foreach ($config["profiles"] AS &$profile) {
    $profile = array_merge([
        "id" => 0,
        "name" => "Profil",
        "minTemp" => 12.0,
        "maxTemp" => 23.0,
        "step" => 0.5,
        "startColor" => 0x0000FF,
        "endColor" => 0xFF0000,
        "scheduleGroups" => [31,96],
    ], $profile);
}

$selfObj = IPS_GetObject($IPS_SELF);
$idInstance = $selfObj['ParentID'];
$instanceObj = IPS_GetObject($idInstance);
$idStatus = CreateVariableByIdent($idInstance,"status","Status", 3, 30,"~TextBox");
// Timerevent Statusvariable leeren
if($_IPS['SENDER'] == "TimerEvent") {
    SetValueString($idStatus,"");
    IPS_SetScriptTimer($_IPS['SELF'], 0);
    return;
}

// Wenn nicht konfiguriert, IPS - ID des Thermostates aus erster passender HM-Instanz auf aktueller Ebene ermitteln
if (!array_key_exists("idHeatingControl",$config) || $config["idHeatingControl"] == null) {
    $config["idHeatingControl"] = null;
    foreach (IPS_GetChildrenIDs($idInstance) AS $idChild) {
        $childObj = IPS_GetObject($idChild);
        if ($childObj === false) continue;
        if ($childObj["ObjectType"] != 1) continue;
        $childInstance = IPS_GetInstance($idChild);
        
        if ($childInstance["InstanceStatus"] != 102) continue;
        if ($childInstance["ModuleInfo"]["ModuleID"] != "{EE4A81C6-5C90-4DB7-AD2F-F6BBD521412E}") continue;
        if (@IPS_GetObjectIDByIdent("ACTUAL_TEMPERATURE", $idChild) !== false) {
            $config["idHeatingControl"] = $idChild;
            break;
        }   
    }
    if ($config["idHeatingControl"] == null)  setStatus("Heating Control Instanz nicht in aktueller Instanz gefunden!","error");
}

setStatus("Stelle Verbindung zu HomeMatic her");

// Daten Heating Control in IPS ermitteln
$heatingControlInstance = @IPS_GetInstance($config["idHeatingControl"]);
if ($heatingControlInstance === false) setStatus("Heating Control Instanz nicht gefunden!","error");
$ipsHeatingDeviceConfig = json_decode(IPS_GetConfiguration($config["idHeatingControl"]));
$config["HMProtocol"] = ["RF","WR","IP"][$ipsHeatingDeviceConfig->Protocol];
if (!array_key_exists("HMAdressHeatingControl",$config) || $config["HMAdressHeatingControl"] == null) $config["HMAdressHeatingControl"] = $ipsHeatingDeviceConfig->Address;
$config["isHMGroup"] = (strpos($config["HMAdressHeatingControl"],"INT") === 0);

// HomeMatic Socket
$HMSocketConfig = json_decode(IPS_GetConfiguration($heatingControlInstance["ConnectionID"]));
if ($config["isHMGroup"] === true) {
    $url = ($HMSocketConfig->UseSSL === true) ? "https://". $HMSocketConfig->Host . ":49292/groups" : "http://". $HMSocketConfig->Host . ":9292/groups";
} else {
    if ($HMSocketConfig->{$config["HMProtocol"] . "Open"} === false) setStatus("HomeMatic Socket ist geschlossen","error");
    $url = ($HMSocketConfig->UseSSL === true) ? "https://".$HMSocketConfig->Host.":".$HMSocketConfig->{$config["HMProtocol"] . "SSLPort"} : "http://".$HMSocketConfig->Host.":".$HMSocketConfig->{$config["HMProtocol"] . "Port"};
}


$xmlrpcClient = new xmlrpc_client($url);
if ($xmlrpcClient === false) setStatus("Keine Verbindung zu HomeMatic","error");


// Änderung von Variablen senden (ausser Profile)
if (in_array($_IPS['SENDER'],["WebFront","VoiceControl"])) {
    $senderObj = IPS_GetObject($_IPS['VARIABLE']);
    if ($senderObj === false) setStatus("Sender Object nicht gefunden!","error");
    if ($senderObj["ObjectType"] === 2) {
        $senderVariable = IPS_GetVariable($_IPS['VARIABLE']);
        switch ($senderObj["ObjectIdent"]) {
            case "BOOST_MODE":
                $response = setValueHeatingControl($senderObj["ObjectIdent"], $_IPS['VALUE'],"boolean",$config);
                setStatus($senderObj["ObjectName"] . " " . (($_IPS['VALUE'] === true) ? "ein" : "aus") . " : " . (($response === true) ? "erfolgreich" : "Fehler") );
                break;
            case "ACTIVE_PROFILE":
            case "SET_POINT_MODE":
                $parameter = ($senderObj["ObjectIdent"] == "SET_POINT_MODE") ? "CONTROL_MODE" : ($senderObj["ObjectIdent"]);
                $valueName = getProfileValueNameFromIPSVariable($_IPS['VARIABLE'],$_IPS['VALUE']);
                if ($valueName === false) setStatus("Der Wert entspricht nicht dem Variablenprofil","error");
                $response = setValueHeatingControl($parameter, $_IPS['VALUE'],"int",$config);
                setStatus("Set " . $parameter . " " . $valueName . " : " . (($response === true) ? "erfolgreich" : "Fehler") );
                break;   
            case "SET_POINT_TEMPERATURE":
                $valueName = getProfileValueNameFromIPSVariable($_IPS['VARIABLE'],$_IPS['VALUE']);
                if ($valueName === false) setStatus("Der Wert entspricht nicht dem Variablenprofil","error");
                // Ist das Fenster offen?
                $idWindowState = @IPS_GetObjectIDByIdent("WINDOW_STATE", $config["idHeatingControl"]);
                if ($idWindowState !== false && GetValueInteger($idWindowState) != 0) {
                    if (!array_key_exists("allowWindowOpenTemperatureChange", $config) || $config["allowWindowOpenTemperatureChange"] === false) setStatus("Die Solltemperatur kann bei geöffnetem Fenster nicht verändert werden.","error");
                    // Änderungen der TEMPERATURE_WINDOW_OPEN wirken sich leider erst beim nächsten Öffnen des Fensters aus
                    $response = setMasterValueHeatingControl("TEMPERATURE_WINDOW_OPEN", $_IPS['VALUE'],"double",$config);
                    setStatus("Set TEMPERATURE_WINDOW_OPEN : " . (($response === true) ? "erfolgreich" : "Fehler") );
                    break;
                }
                // Es wird die aktuelle Solltemperatur verändert. Bei Profilen bis zum nächsten Umschalten, beim manuellen Modus für diesen dauerhaft
                $response = setValueHeatingControl("SET_POINT_TEMPERATURE", $_IPS['VALUE'],"double",$config);
                setStatus("Set " . $senderObj["ObjectName"] . " " . $valueName . " : " . (($response === true) ? "erfolgreich" : "Fehler") );
                break;
            case "HEATING_MODE_COMBINED":
                switch ($_IPS['VALUE']) {
                    case 10: 
                        setValueHeatingControl("CONTROL_MODE", 1,"int",$config);
                        break;
                    case 11: 
                        //setValueHeatingControl("CONTROL_MODE", 2,"int",$config);
                        setStatus("Nicht möglich: Für den Partymode müssen Start- und Endezeitpunkt konfiguiert werden.");
                        return;
                        break;
                    case 1:
                    case 2:
                    case 3:
                        setValueHeatingControl("CONTROL_MODE", 0,"int",$config);
                        setValueHeatingControl("ACTIVE_PROFILE", (int)$_IPS['VALUE'],"int",$config);
                        break;
                }
                break;
            default:
                setStatus("Sendung von unbekannter Variable: " . $_IPS['VARIABLE'] . " / " . $senderObj["ObjectIdent"] ,"error");
        }
        return;
    }
}


$action = (array_key_exists("targetAction",$_IPS)) ? $_IPS["targetAction"] : "install";

switch ($action) {
    // Button-Lock bei geöffnetem Fenster setzen
    case "toggleWindowState":
        if (!array_key_exists("lockDeviceButtonOnOpenedWindow",$config) || $config["lockDeviceButtonOnOpenedWindow"] === false) return;
        $idWindowState = @IPS_GetObjectIDByIdent("WINDOW_STATE", $config["idHeatingControl"]);
        if ($idWindowState === false) setStatus("ID für WINDOW_STATE konnte nicht ermittelt werden","error");
        $response = setMasterValueHeatingControl("GLOBAL_BUTTON_LOCK", (GetValueInteger($idWindowState) != 0),"boolean",$config,0);
        setStatus("Set BUTTON_LOCK : " . (($response === true) ? "erfolgreich" : "Fehler") );
        return;
        break;
}

$deviceData = getDeviceDataFromHomeMatic($config["HMAdressHeatingControl"]);

// Namen der Profile aus Variablenprofile für ACTIVE_PROFILE ermitteln
$idActiveProfile = IPS_GetObjectIDByIdent("ACTIVE_PROFILE", $config["idHeatingControl"]);
if ($idActiveProfile !== false) {
    foreach($config["profiles"] as &$profile) {
        $name = getProfileValueNameFromIPSVariable($idActiveProfile,$profile["id"]);
        if ($name !== false) $profile["name"] = $name;
    }
}

setStatus("Lade Daten");

// IPS-Events erstellen und befüllen
foreach ($deviceData["profiles"] as $profileKey => $profileValues) {
    if (!array_key_exists($profileKey, $config["profiles"])) continue;
    createIPSEventForProfile($idInstance,$config["profiles"][$profileKey]);
    if ($action == "install" || $action == "loadProfiles") setIPSEventScheduleGroupsForProfile($idInstance,$config["profiles"][$profileKey],$profileValues["weekdays"]); 
    if ($action == "storeProfiles") $deviceData["profiles"][$profileKey]["weekdays"] = getProfileDataFromIPSEvent($idInstance,$config["profiles"][$profileKey]["id"],$deviceData["profiles"][$profileKey]["itemsPerWeekday"]);
}

setStatus("Daten geladen");

if ($action == "storeProfiles") {
    setStatus("Speichere Wochenprofile");
    storeProfilesInHomeMatic($config["HMAdressHeatingControl"],$deviceData);
    setStatus("Wochenprofile gespeichert");
}

if ($action == "install") {
    initializeInstance($idInstance,$config); 
}

// Setzen Status-Variable
function setStatus($message,$severity = "info") {
    global $idStatus;
    SetValueString($idStatus,$message);
    IPS_SetScriptTimer($_IPS['SELF'], 10);
    if ($severity == "error") die($message);
}


// *** HomeMatic XMLRPC Kommunikation

// Konfiguration von HM-Device via XMLRPC holen
function getDeviceDataFromHomeMatic($HMAdress) {
    $request = new xmlrpcmsg("getDeviceDescription",[new xmlrpcval("$HMAdress", "string"), new xmlrpcval("MASTER", "string")]);
	$deviceDescription = xmlrpcSend($request);
    if ($deviceDescription === false) setStatus("Fehler beim Lesen der Gerätebeschreibung","error");
    if ($deviceDescription["TYPE"] != "HEATING_CLIMATECONTROL_TRANSCEIVER") setStatus("Gerät ist nicht vom Typ HEATING_CLIMATECONTROL_TRANSCEIVER","error");
    $data = [
        "deviceType" => $deviceDescription["PARENT_TYPE"],
        "profiles" => [],
    ];

    $params = xmlrpcReadParamSet($HMAdress);
    if ($params === false) setStatus("Fehler beim Lesen der Gerätedaten","error");
    
    foreach ($params AS $paramName => $paramValue) {
        $paramNameArr = explode("_",$paramName);
        if (! preg_match("/^P[1-9]$/", $paramNameArr[0])) {
            $data[$paramName] = $paramValue;
            continue;
        }
        $data["profiles"][$paramNameArr[0]]["weekdays"][$paramNameArr[2]][$paramNameArr[3]][$paramNameArr[1]] = $paramValue;
    }
    /* 
    ENDTIME sind Minuten seit 0:00 Uhr in 10-Minuten-Schritten
    Es gibt fest 13 Speicherplätze pro Profil / Tag
    Nach 24:00 Uhr ist der Inhalt der ebenfalls übermittelten freien unbelegten Speicherplätze uninteressant -> ENDTIME ist immer 14400 -> 24:00 Uhr
    */
    foreach ($data["profiles"] AS $profile => $profileValues) {
        foreach ($data["profiles"][$profile]["weekdays"] AS $weekday => $weekdayValues) {
            $data["profiles"][$profile]["itemsPerWeekday"] = count($weekdayValues);
            $endtime = 0;
            for ($idx = 1; $idx <= $data["profiles"][$profile]["itemsPerWeekday"]; $idx++) {
                if ($endtime == 1440) {
                    unset($data["profiles"][$profile]["weekdays"][$weekday][$idx]);
                    continue;
                }
                $endtime = $data["profiles"][$profile]["weekdays"][$weekday][$idx]["ENDTIME"];
                $data["profiles"][$profile]["weekdays"][$weekday][$idx]["time"] = sprintf("%02d:%02d", floor($endtime/60), $endtime%60); 
            }
            ksort($data["profiles"][$profile]["weekdays"][$weekday]);
        }
    }
    ksort($data["profiles"]);
    return $data;
}

// Wochenprofile an HomeMatic via XMLRPC übermitteln
function storeProfilesInHomeMatic($HMAdress, $deviceData) {
    $sendParams = new xmlrpcval();
    foreach ($deviceData["profiles"] AS $profileName => $profileData) {
        foreach ($profileData["weekdays"] AS $weekdayName => $weekdayData) {
            if (count($weekdayData) == 0) continue;
            foreach ($weekdayData AS $idx => $pointData) {
                if ($idx > $profileData["itemsPerWeekday"]) continue;
                $key = $profileName . "_TEMPERATURE_" . $weekdayName . "_" . $idx;
                $param = array($key => new xmlrpcval($pointData["TEMPERATURE"], "double"));
		        $sendParams->addStruct($param);
                $key = $profileName . "_ENDTIME_" . $weekdayName . "_" . $idx;
                $param = array($key => new xmlrpcval($pointData["ENDTIME"], "int"));
		        $sendParams->addStruct($param);
            }
            // Auffüllen
            for ($i = count($weekdayData) + 1; $i <= $profileData["itemsPerWeekday"]; $i++) {
                $key = $profileName . "_TEMPERATURE_" . $weekdayName . "_" . $i;
                $param = array($key => new xmlrpcval($pointData["TEMPERATURE"], "double"));
		        $sendParams->addStruct($param);
                $key = $profileName . "_ENDTIME_" . $weekdayName . "_" . $i;
                $param = array($key => new xmlrpcval($pointData["ENDTIME"], "int"));
		        $sendParams->addStruct($param);
            }
        } 
    }
    $request = new xmlrpcmsg("putParamset",	[new xmlrpcval($HMAdress, "string"), new xmlrpcval("MASTER", "string"), $sendParams]);
	$response = xmlrpcSend($request);
    return true;
}

// Liest die Parameter von einem Gerät via XMLRPC aus.
function xmlrpcReadParamSet($HMAdress, $key= "MASTER", $param = false) {
	// Build request
	$request = new xmlrpcmsg("getParamset",[new xmlrpcval($HMAdress, "string"), new xmlrpcval($key, "string")]);
	$messages = xmlrpcSend($request);
	//var_dump($messages);
	if ($param !== false) return $messages[$param];
	return $messages;
}


// Senden XMLPRC
function xmlrpcSend($request) {
    global $xmlrpcClient;
	$response = $xmlrpcClient->send($request);
    if ( $response->errno == 0 ) {
        $messages = php_xmlrpc_decode($response->value());
    } else {
        //print_r($response);
        setStatus("Kommunikationsfehler: $response->errno / $response->errstr","error");
    }
	return $messages;
}

// Value in Heating Control Instanz entweder über Instanz in IPS oder bei Gruppen mittels XMLRPC setzen
function setValueHeatingControl($key, $value, $type, &$config) {
    if ($config["isHMGroup"] === true) {       
        $request = new xmlrpcmsg("setValue",[new xmlrpcval($config["HMAdressHeatingControl"], "string"), new xmlrpcval($key, "string"), new xmlrpcval($value, $type)]);
	    $response = xmlrpcSend($request);
        return true;
    } 
    switch ($type) {
        case "boolean":
            return HM_WriteValueBoolean($config["idHeatingControl"], $key, $value);
        case "double":
            return HM_WriteValueFloat($config["idHeatingControl"], $key, $value);  
        case "int":
            return HM_WriteValueInteger($config["idHeatingControl"], $key, $value);
        case "string":
            return HM_WriteValueString($config["idHeatingControl"], $key, $value);
    }

    return false;
}

// Setzen eines Values Heating Control Instanz in DeviceConfig (Set Master) via XMLRPC
function setMasterValueHeatingControl($key, $value, $type, &$config, $channel = 1) {
	$HMAdress = str_replace(":1",":",$config["HMAdressHeatingControl"]) . $channel;
    $sendParams = new xmlrpcval();
	$sendParams->addStruct([$key => new xmlrpcval($value, $type)]);
    // Build request
	$request = new xmlrpcmsg("putParamset",	[new xmlrpcval($HMAdress, "string"), new xmlrpcval("MASTER", "string"), $sendParams]);
	$messages = xmlrpcSend($request);

	return true;
}

// *** IPS Wochenplan

// Wochenplan in IPS anlegen
function createIPSEventForProfile(&$idInstance,&$params) {
    $idEvent = @IPS_GetObjectIDByIdent("profile_" . $params["id"], $idInstance);
    if ($idEvent !== false) {
        // Namen prüfen und ggf anpassen
        $eventObj = IPS_GetObject($idEvent);
        if ( $eventObj["ObjectName"] != $params["name"]) IPS_SetName($idEvent, $params["name"]);
        return true;
    }

    $steps = ($params["maxTemp"] - $params["minTemp"]) / $params["step"];
    $colors = colorTransition( $params["startColor"],$params["endColor"],$steps);
    $idEvent = IPS_CreateEvent(2);
    IPS_SetParent($idEvent, $idInstance);
    //IPS_SetPosition($idEvent, 20 + $params["id"]);
    IPS_SetIdent($idEvent,"profile_" . $params["id"]);
    // IPS-App hat Problem, wenn ID nicht bei 1 beginnt
    for ($i = 0; $i <= $steps; $i++) {
        $temperature = $params["minTemp"]  + ($i * $params["step"]);
        if ($params["step"] < 1) $temperature = number_format($temperature,1);
        IPS_SetEventScheduleAction($idEvent, $i +1, $temperature . "°", $colors[$i], "");
    }
    for ($i = 0; $i < count($params["scheduleGroups"]); $i++) {
        IPS_SetEventScheduleGroup($idEvent, $i, $params["scheduleGroups"][$i]); 
    }
    IPS_SetName($idEvent, $params["name"]);
    return true;
}

// Wochenplan befüllen
function setIPSEventScheduleGroupsForProfile($idInstance,$params,$eventData) {
    $idEvent = @IPS_GetObjectIDByIdent("profile_" . $params["id"], $idInstance);

    if ($idEvent === false) setStatus("Profil nicht gefunden","error");
    $eventObj = IPS_GetEvent($idEvent);
    if ($eventObj == false) setStatus("Event nicht gefunden","error");
    if ($eventObj["EventType"] !== 2) setStatus("Falscher Eventtyp","error");
    
    // Reset
    foreach ($eventObj["ScheduleGroups"] AS $scheduleGroup) {
        IPS_SetEventScheduleGroup($idEvent, $scheduleGroup["ID"], 0);
    }

    // Mapping AktionsId zu Temperatur
    $actionIdMapping = [];
    foreach ($eventObj["ScheduleActions"] AS $scheduleAction) {
        $actionIdMapping[(float)str_replace("°","",$scheduleAction["Name"])] = $scheduleAction["ID"];
    }
    // Wochentage in Reihenfolge Profileinstellung HomeMatic
    $weekDays = [
        6 => "SATURDAY",
        7 => "SUNDAY",
        1 => "MONDAY",
        2 => "TUESDAY",
        3 => "WEDNESDAY",
        4 => "THURSDAY",
        5 => "FRIDAY",
    ];
    for ($i = 0; $i < count($params["scheduleGroups"]); $i++) {
        IPS_SetEventScheduleGroup($idEvent, $i, $params["scheduleGroups"][$i]);
        $weekday = null;
        // erster enthaltener Wochentag
        foreach ($weekDays AS $weekDayIdx => $weekDayName) {
            if (($params["scheduleGroups"][$i] & pow(2,$weekDayIdx-1)) > 0) {
                $weekday = $weekDayName;
                break;
            }
        }
        
        if ($weekday == null) continue;
        if (!array_key_exists($weekday,$eventData)) continue;
        $startTime = [0,0];
        foreach ($eventData[$weekday] AS $pointIdx => $pointData) {
            if ($pointData["TEMPERATURE"] < $params["minTemp"]) $pointData["TEMPERATURE"] = $params["minTemp"];
            if ($pointData["TEMPERATURE"] > $params["maxTemp"]) $pointData["TEMPERATURE"] = $params["maxTemp"];
            $scheduleActionId = (int) (($pointData["TEMPERATURE"] - $params["minTemp"]) / $params["step"]) + 1;
            IPS_SetEventScheduleGroupPoint($idEvent, $i, $pointIdx, (int)$startTime[0], (int)$startTime[1], 0, $scheduleActionId);
            $startTime = explode(":",$pointData["time"]); 
        }
        // Workarround: in IPSView wird letzter Eintrag nur bis 23:00 dargestellt -> erzeuge Dummyeintrag ab 23:59
        $scheduleActionId = ($scheduleActionId == 1) ? 2 : $scheduleActionId -1;
        IPS_SetEventScheduleGroupPoint($idEvent, $i, $pointIdx + 1, 23, 59, 0, $scheduleActionId);
    }

}

// Farbwerte für Wochenplan ermitteln (Verlauf zwischen zwei Farben)
function colorTransition($theColorBegin = 0x0000FF,  $theColorEnd = 0xFF0000,$theNumSteps = 10) {
    $theColorBegin = (($theColorBegin >= 0x000000) && ($theColorBegin <= 0xffffff)) ? $theColorBegin : 0x000000;
    $theColorEnd = (($theColorEnd >= 0x000000) && ($theColorEnd <= 0xffffff)) ? $theColorEnd : 0xffffff;
    $theNumSteps = (($theNumSteps > 0) && ($theNumSteps < 256)) ? $theNumSteps : 16;

    $theR0 = ($theColorBegin & 0xff0000) >> 16;
    $theG0 = ($theColorBegin & 0x00ff00) >> 8;
    $theB0 = ($theColorBegin & 0x0000ff) >> 0;

    $theR1 = ($theColorEnd & 0xff0000) >> 16;
    $theG1 = ($theColorEnd & 0x00ff00) >> 8;
    $theB1 = ($theColorEnd & 0x0000ff) >> 0;
    $colors = [];
    for ($i = 0; $i <= $theNumSteps; $i++) {
        $theR = interpolate($theR0, $theR1, $i, $theNumSteps);
        $theG = interpolate($theG0, $theG1, $i, $theNumSteps);
        $theB = interpolate($theB0, $theB1, $i, $theNumSteps);
        $colors[] = ((($theR << 8) | $theG) << 8) | $theB;
    }
    return $colors;
}

// return the interpolated color value between pBegin and pEnd
function interpolate($pBegin, $pEnd, $pStep, $pMax) {
    if ($pBegin < $pEnd) {
        return (($pEnd - $pBegin) * ($pStep / $pMax)) + $pBegin;
    } else {
        return (($pBegin - $pEnd) * (1 - ($pStep / $pMax))) + $pEnd;
    }
}

// Daten aus Wochenplan auslesen
function getProfileDataFromIPSEvent($idInstance,$idEvent,$maxPoints) {
    $idEvent = @IPS_GetObjectIDByIdent("profile_" . $idEvent, $idInstance);
    if ($idEvent === false) setStatus("Profil nicht gefunden","error");
    $event = IPS_GetEvent($idEvent);
    if ($event == false) setStatus("Event nicht gefunden","error");

    $weekDays = [
        6 => "SATURDAY",
        7 => "SUNDAY",
        1 => "MONDAY",
        2 => "TUESDAY",
        3 => "WEDNESDAY",
        4 => "THURSDAY",
        5 => "FRIDAY",
    ];
    // Mapping AktionsId zu Temperatur
    $actionIdMapping = [];
    foreach ($event["ScheduleActions"] AS $scheduleAction) {
        $actionIdMapping[$scheduleAction["ID"]] = (float) str_replace("°","",$scheduleAction["Name"]);
    }

    $data = [];
    foreach($event["ScheduleGroups"] AS $scheduleGroup) {
        foreach ($weekDays AS $weekDayIdx => $weekDayName) {
            if (($scheduleGroup["Days"] & pow(2,$weekDayIdx-1)) > 0) {
                $data[$weekDayName] = [];
                $i = 1;
                foreach($scheduleGroup["Points"] AS $point) {
                    if ($i == $maxPoints) continue;
                    // letzen Eintrag mit 23:59 (IPSView Workarround) ignorieren
                    if ($point["Start"]["Hour"] == 23 && $point["Start"]["Minute"] == 59) continue;

                    if (($point["Start"]["Hour"] + $point["Start"]["Minute"] + $point["Start"]["Second"]) != 0) {
                        $point["Start"]["Minute"] = round($point["Start"]["Minute"], -1, PHP_ROUND_HALF_DOWN);
                        $data[$weekDayName][$i] = [
                            "ENDTIME" => (int) (((int)$point["Start"]["Hour"] * 60) + (int)$point["Start"]["Minute"]),
                            "TEMPERATURE" => $temperature,
                            "time" => sprintf("%02d:%02d", $point["Start"]["Hour"], $point["Start"]["Minute"]),
                        ];
                        $i++;
                    }
                    $temperature = $actionIdMapping[$point["ActionID"]];
                }
                $data[$weekDayName][$i] = [
                    "ENDTIME" => 1440,
                    "TEMPERATURE" => $temperature,
                    "time" => "24:00",
                ];
            }
        }
    }
    return $data;

}

// *** Initalisierung der Elemente der Instanz

function initializeInstance($idInstance,&$config) {
    if (!is_array($config) || !array_key_exists("elements",$config) || !is_array($config["elements"])) setStatus("Die Konfiguration ist fehlerhaft","error");
    $position = 10;
    foreach ($config["elements"] AS $idx => $element) {
        if (!array_key_exists("type",$element)) continue;
        $position++;
        switch ($element["type"]) {
            case "linkHeatingControlVariable":
            case "setHeatingControlProfile":
                if (!array_key_exists("sourceIdent",$element)) setStatus("Parameter sourceIdent für Element $idx nicht vorhanden","error");
                $idSource = @IPS_GetObjectIDByIdent($element["sourceIdent"], $config["idHeatingControl"]);
                if ($idSource === false) continue 2;
                $sourceVariable = IPS_GetVariable($idSource);
                // Variablenprofil zuweisen (nur wenn leer)
                if (array_key_exists("profile",$element) && $element["profile"] != null && ($sourceVariable["VariableCustomProfile"] == "" || $config["overwriteExistingVariablesProfiles"] === true)) {
                    if (CreateVariableProfileByName($element["profile"]) === false) setStatus("Variablenprofil " . $element["profile"] . " existiert nicht","error");
                    if (IPS_SetVariableCustomProfile($idSource, $element["profile"]) === false) setStatus("Variablenprofil " . $element["profile"] . " konnte nicht zugewiesen werden","error");
                }
                // Eigene ScriptId als CustomAction setzen
                if (array_key_exists("selfAsCustomAction",$element) && $element["selfAsCustomAction"] === true) {
                    if (IPS_SetVariableCustomAction($idSource, $_IPS['SELF'])=== false) setStatus("CustomAction konnte nicht zugewiesen werden","error");
                }
                 // OnChange Trigger
                if (array_key_exists("triggerOnChange",$element)) {
                    if (@IPS_GetObjectIDByIdent("triggerOnChange", $idSource) === false) {
                        $idTrigger = IPS_CreateEvent (0);
                        IPS_SetIdent ($idTrigger, "triggerOnChange");
                        IPS_SetParent($idTrigger, $idSource);
                        IPS_SetEventTrigger ($idTrigger, 1, $idSource);
                        IPS_SetEventScript ($idTrigger, $element["triggerOnChange"]);
                        IPS_SetEventActive ($idTrigger, true);
                    }
                }

                if ($element["type"] == "setHeatingControlProfile") break;
                $linkIdent = "link_" . $element["sourceIdent"];
                $idLink = @IPS_GetObjectIDByIdent($linkIdent, $idInstance);
                if ($idLink !== false) continue 2;
                $idLink = IPS_CreateLink ();
                IPS_SetParent($idLink, $idInstance);
                IPS_SetLinkTargetID ($idLink, $idSource);
                if (!array_key_exists("name",$element)) $element["name"] = $sourceVariable["ObjectName"];
                IPS_SetName( $idLink, $element["name"]);
                IPS_SetIdent( $idLink, $linkIdent);
                IPS_SetPosition( $idLink, $position);
                break;
            case "linkHeatingControlMaintenanceVariables":
                if (!array_key_exists("sourceIdent",$element)) setStatus("Parameter sourceIdent für Element $idx nicht vorhanden","error");
                foreach (IPS_GetChildrenIDs($idInstance) AS $idChild) {
                    $childObj = IPS_GetObject($idChild);
                    if ($childObj === false) continue;
                    if ($childObj["ObjectType"] != 1) continue;
                    $childInstance = IPS_GetInstance($idChild);
                    if ($childInstance["InstanceStatus"] != 102) continue;
                    if ($childInstance["ModuleInfo"]["ModuleID"] != "{EE4A81C6-5C90-4DB7-AD2F-F6BBD521412E}") continue;
                    $idSource = @IPS_GetObjectIDByIdent($element["sourceIdent"], $idChild);
                    if ($idSource !== false) {
                        $position++;
                        $sourceVariable = IPS_GetVariable($idSource);
                        // Variablenprofil zuweisen (nur wenn leer)
                        if (array_key_exists("profile",$element) && $element["profile"] != null && ($sourceVariable["VariableCustomProfile"] == "" || $config["overwriteExistingVariablesProfiles"] === true)) {
                            if (CreateVariableProfileByName($element["profile"]) === false) setStatus("Variablenprofil " . $element["profile"] . " existiert nicht","error");
                            if (IPS_SetVariableCustomProfile($idSource, $element["profile"]) === false) setStatus("Variablenprofil " . $element["profile"] . " konnte nicht zugewiesen werden","error");
                        }
                        $linkIdent = "link_" . $idChild . $element["sourceIdent"];
                        $idLink = @IPS_GetObjectIDByIdent($linkIdent, $idInstance);
                        if ($idLink !== false) continue;
                        $idLink = IPS_CreateLink ();
                        IPS_SetParent($idLink, $idInstance);
                        IPS_SetLinkTargetID ($idLink, $idSource);
                        if (!array_key_exists("name",$element)) $element["name"] = $sourceVariable["ObjectName"];
                        IPS_SetName( $idLink, str_replace(" - Wartung","",$childObj["ObjectName"]) . " - " . $element["name"]);
                        IPS_SetIdent( $idLink, $linkIdent);
                        IPS_SetPosition( $idLink, $position);
                        if ($childObj["ParentID"] == $idInstance) {
                            IPS_SetPosition($idChild, 91);
                            IPS_SetHidden ($idChild, true);
                        }
                    }
                }
                break;
            case "heatingModeCombined":
                $varIdent = "HEATING_MODE_COMBINED";
                $idVar = @IPS_GetObjectIDByIdent($varIdent, $idInstance);
                if ($idVar !== false) continue 2;
                $idVar = IPS_CreateVariable(1);
                IPS_SetParent($idVar, $idInstance);
                if (!array_key_exists("name",$element)) $element["name"] = "Heizmodus";
                IPS_SetName( $idVar, $element["name"]);
                IPS_SetIdent( $idVar, $varIdent);
                IPS_SetPosition( $idVar, $position);
                if (array_key_exists("profile",$element) && $element["profile"] != null) {
                    if (CreateVariableProfileByName($element["profile"]) === false) setStatus("Variablenprofil " . $element["profile"] . " existiert nicht","error");
                    if (IPS_SetVariableCustomProfile($idVar, $element["profile"]) === false) setStatus("Variablenprofil " . $element["profile"] . " konnte nicht zugewiesen werden","error");
                }
                // Eigene ScriptId als CustomAction setzen
                if (array_key_exists("selfAsCustomAction",$element) && $element["selfAsCustomAction"] === true) {
                    if (IPS_SetVariableCustomAction($idVar, $_IPS['SELF'])=== false) setStatus("CustomAction konnte nicht zugewiesen werden","error");
                }
                // Initial setzen
                switch (GetValueInteger(IPS_GetObjectIDByIdent("SET_POINT_MODE", $config["idHeatingControl"]))) {
                    case 0:
                        SetValueInteger($idVar,GetValueInteger(IPS_GetObjectIDByIdent("ACTIVE_PROFILE", $config["idHeatingControl"])));
                        break;
                    case 1:
                        SetValueInteger($idVar,10);
                        break;
                     case 2:
                        SetValueInteger($idVar,11);
                        break;
                }
                // Trigger
                $idTrigger = IPS_CreateEvent (0);
                IPS_SetParent($idTrigger, $idVar);
                IPS_SetEventTrigger ($idTrigger, 0, IPS_GetObjectIDByIdent("SET_POINT_MODE", $config["idHeatingControl"]));
                IPS_SetEventScript ($idTrigger, '<?php
                    if ($_IPS["VALUE"] == 0) SetValueInteger(' . $idVar . ',GetValueInteger(' . IPS_GetObjectIDByIdent("ACTIVE_PROFILE", $config["idHeatingControl"]) . '));
                    if ($_IPS["VALUE"] == 1) SetValueInteger(' . $idVar . ',10);
                    if ($_IPS["VALUE"] == 2) SetValueInteger(' . $idVar . ',11);
                ');
                IPS_SetEventActive ($idTrigger, true);
                break;
            case "script":
                if (!array_key_exists("ident",$element)) setStatus("Parameter ident für Element $idx nicht vorhanden","error");
                $idScript = @IPS_GetObjectIDByIdent($element["ident"], $idInstance);
                if ($idScript !== false) continue 2;
                $idScript = IPS_CreateScript(0);
                IPS_SetParent($idScript, $idInstance);
                if (!array_key_exists("name",$element)) $element["name"] = "Script $idx";
                IPS_SetName($idScript, $element["name"]);
                IPS_SetIdent($idScript, $element["ident"]);
                if (array_key_exists("content",$element)) IPS_SetScriptContent($idScript, $element["content"]);
                IPS_SetPosition($idScript, $position);
                break;
            case "profiles":
                foreach (IPS_GetChildrenIDs($idInstance) AS $idChild) {
                    $childObj = IPS_GetObject($idChild);
                    $startPosition = $position;
                    if (strpos($childObj["ObjectIdent"],"profile_") === 0) {
                        IPS_SetPosition($idChild, $startPosition + (int)substr($childObj["ObjectIdent"],-1) - 1);
                        $position++;
                    }
                    $position++;
                }
                break;
            case "setPosition":
                if (!array_key_exists("ident",$element)) setStatus("Parameter ident für Element $idx nicht vorhanden","error");
                $idElement = @IPS_GetObjectIDByIdent($element["ident"], $idInstance);
                if ($idElement === false) continue 2;
                IPS_SetPosition($idElement, $position);
                break;
        }
       
    }
    $heatingObj = IPS_GetObject($config["idHeatingControl"]);
    if ($heatingObj !== false && $heatingObj["ParentID"] == $idInstance) {
        IPS_SetPosition($config["idHeatingControl"], 90);
        IPS_SetHidden ($config["idHeatingControl"], true);
    }
    IPS_SetPosition($_IPS['SELF'], 99);
    IPS_SetHidden ($_IPS['SELF'], true);

}

// Variablenprofile anlegen
function CreateVariableProfileByName($name) {
    if (IPS_VariableProfileExists($name) === true) return true;
    switch ($name) {
        case "MSC.HMIPHeatingControl.HeatingModeCombined":
            IPS_CreateVariableProfile("MSC.HMIPHeatingControl.HeatingModeCombined", 1);
            IPS_SetVariableProfileText("MSC.HMIPHeatingControl.HeatingModeCombined", "", "");
            IPS_SetVariableProfileValues("MSC.HMIPHeatingControl.HeatingModeCombined", 0, 11, 0);
            IPS_SetVariableProfileDigits("MSC.HMIPHeatingControl.HeatingModeCombined", 0);
            IPS_SetVariableProfileIcon("MSC.HMIPHeatingControl.HeatingModeCombined", "Radiator");
            IPS_SetVariableProfileAssociation("MSC.HMIPHeatingControl.HeatingModeCombined", 1, "Profil 1", "", 32960);
            IPS_SetVariableProfileAssociation("MSC.HMIPHeatingControl.HeatingModeCombined", 2, "Profil 2", "", 32960);
            IPS_SetVariableProfileAssociation("MSC.HMIPHeatingControl.HeatingModeCombined", 3, "Profil 3", "", 32960);
            IPS_SetVariableProfileAssociation("MSC.HMIPHeatingControl.HeatingModeCombined", 10, "Manuell", "", 32960);
            IPS_SetVariableProfileAssociation("MSC.HMIPHeatingControl.HeatingModeCombined", 11, "Party", "", 32960);
            break;
        case "MSC.HMIPHeatingControl.SetPointMode":
            IPS_CreateVariableProfile("MSC.HMIPHeatingControl.SetPointMode", 1);
            IPS_SetVariableProfileText("MSC.HMIPHeatingControl.SetPointMode", "", "");
            IPS_SetVariableProfileValues("MSC.HMIPHeatingControl.SetPointMode", 0, 0, 0);
            IPS_SetVariableProfileDigits("MSC.HMIPHeatingControl.SetPointMode", 0);
            IPS_SetVariableProfileIcon("MSC.HMIPHeatingControl.SetPointMode", "");
            IPS_SetVariableProfileAssociation("MSC.HMIPHeatingControl.SetPointMode", 0, "Automatisch", "", 32960);
            IPS_SetVariableProfileAssociation("MSC.HMIPHeatingControl.SetPointMode", 1, "Manuell", "", 32960);
            IPS_SetVariableProfileAssociation("MSC.HMIPHeatingControl.SetPointMode", 2, "Party", "", 32960);
            break;
        case "MSC.HMIPHeatingControl.Temperature":
            IPS_CreateVariableProfile("MSC.HMIPHeatingControl.Temperature", 2);
            IPS_SetVariableProfileText("MSC.HMIPHeatingControl.Temperature", "", "");
            IPS_SetVariableProfileValues("MSC.HMIPHeatingControl.Temperature", 4.5, 30.5, 0.5);
            IPS_SetVariableProfileDigits("MSC.HMIPHeatingControl.Temperature", 1);
            IPS_SetVariableProfileIcon("MSC.HMIPHeatingControl.Temperature", "");
            IPS_SetVariableProfileAssociation("MSC.HMIPHeatingControl.Temperature", 4.5, "Aus", "", -1);
            // wegen IPSView mit Werten auffüllen und nicht Platzhalter zu Formatierung verwenden
            for ($i = 5; $i <= 30; $i = $i + 0.5) {
                IPS_SetVariableProfileAssociation("MSC.HMIPHeatingControl.Temperature", $i, sprintf("%0.1f °C",$i), "", -1);
            }
            IPS_SetVariableProfileAssociation("MSC.HMIPHeatingControl.Temperature", 30.5, "Ein", "", -1);
            break;
        case "MSC.HMIPHeatingControl.ActiveProfile":
            IPS_CreateVariableProfile("MSC.HMIPHeatingControl.ActiveProfile", 1);
            IPS_SetVariableProfileText("MSC.HMIPHeatingControl.ActiveProfile", "", "");
            IPS_SetVariableProfileValues("MSC.HMIPHeatingControl.ActiveProfile", 0, 0, 0);
            IPS_SetVariableProfileDigits("MSC.HMIPHeatingControl.ActiveProfile", 0);
            IPS_SetVariableProfileIcon("MSC.HMIPHeatingControl.ActiveProfile", "");
            IPS_SetVariableProfileAssociation("MSC.HMIPHeatingControl.ActiveProfile", 1, "Profil 1", "", -1);
            IPS_SetVariableProfileAssociation("MSC.HMIPHeatingControl.ActiveProfile", 2, "Profil 2", "", -1);
            IPS_SetVariableProfileAssociation("MSC.HMIPHeatingControl.ActiveProfile", 3, "Profil 3", "", -1);
            break;
        case "MSC.HMIPHeatingControl.Window":
            IPS_CreateVariableProfile("MSC.HMIPHeatingControl.Window", 1);
            IPS_SetVariableProfileText("MSC.HMIPHeatingControl.Window", "", "");
            IPS_SetVariableProfileValues("MSC.HMIPHeatingControl.Window", 0, 0, 0);
            IPS_SetVariableProfileDigits("MSC.HMIPHeatingControl.Window", 0);
            IPS_SetVariableProfileIcon("MSC.HMIPHeatingControl.Window", "Window");
            IPS_SetVariableProfileAssociation("MSC.HMIPHeatingControl.Window", 0, "Geschlossen", "Window-100", 65280);
            IPS_SetVariableProfileAssociation("MSC.HMIPHeatingControl.Window", 1, " Geöffnet", "Window-0", 16711680);
            break;
        default:
            return false;
    }
    return true;

}

// *** IPS Hilfsfunktionen

function CreateVariableByIdent($id, $ident, $name, $type, $position = 0, $profile = null) {
    $vid = @IPS_GetObjectIDByIdent($ident, $id);
    if($vid !== false) return $vid;
    $vid = IPS_CreateVariable($type);
    IPS_SetParent($vid, $id);
    IPS_SetName($vid, $name);
    IPS_SetIdent($vid, $ident);
    if ($profile !== null) IPS_SetVariableCustomProfile($vid, $profile); 
    if ($position != 0 && is_numeric($position)) IPS_SetPosition($vid, $position);
    return $vid;
}

function getProfileFromIPSVariable($idVariable) {
    $variable = IPS_GetVariable($idVariable);
    if ($variable === false) return false;
    $idProfile = ($variable["VariableCustomProfile"] != "") ? $variable["VariableCustomProfile"] : $variable["VariableProfile"];
    if ($idProfile == "") return false;
    return @IPS_GetVariableProfile($idProfile);
}

function getProfileValueNameFromIPSVariable($idVariable,$value) {
    $profile = getProfileFromIPSVariable($idVariable);
    if ($profile === false) return false;
    $response = false;
    if (in_array($profile["ProfileType"],[1,2]) ) {
        if (($profile["MinValue"] + $profile["MaxValue"]) > 0 && ($value < $profile["MinValue"] || $value > $profile["MaxValue"])) return false;
        $response = GetValueFormattedEx($idVariable,$value);
    }

    foreach ($profile["Associations"] as $association) {
        if ($association["Value"] == $value) return $association["Name"];
    }
    return $response;
}